# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: savu_loader
   :platform: Unix
   :synopsis: A class for loading savu output data.

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""

import os
import json
import h5py
import copy
import numpy as np

import savu.plugins.utils as pu
import savu.core.utils as cu
from savu.plugins.utils import register_plugin
from savu.plugins.loaders.base_loader import BaseLoader

from savu.core.utils import ensure_string


@register_plugin
class SavuNexusLoader(BaseLoader):
    """
    A class to load datasets, and associated metadata, from a Savu output
    nexus file.

    By default, the last instance of each unique dataset name will be loaded.
    Opt instead to load a subset of these datasets, or individual datasets by
    populating the parameters.

    :u*param preview: A slice list of required frames to apply to ALL \
    datasets, else a dictionary of slice lists where the key is the dataset \
    name. Default: {}.

    :u*param datasets: Override the default by choosing specific dataset(s) to\
    load, by stating the NXdata name. Default: [].
    :u*param names: Override the dataset names associated with the datasets \
    parameter above. Default: [].
    """

    def __init__(self, name='SavuNexusLoader'):
        super(SavuNexusLoader, self).__init__(name)
        self._all_preview_params = None

    def setup(self):
        self._all_preview_params = copy.deepcopy(self.parameters['preview'])

        datasets = []
        with h5py.File(self.exp.meta_data.get('data_file'), 'r') as nxsfile:
            datasets = self._read_nexus_file(nxsfile, datasets)
            datasets = self._update_plugin_numbers(datasets)

            exp_dict = self.exp.meta_data.get_dictionary()
            if 'checkpoint_loader' in list(exp_dict.keys()):
                self.__checkpoint_reload(nxsfile, datasets)
            else:
                self.__reload(nxsfile, datasets)

    def __reload(self, nxsfile, datasets):
        if self.parameters['datasets']:
            datasets = self.__get_parameter_datasets(datasets)
        else:
            datasets = self._last_unique_datasets(datasets)
        self._create_datasets(nxsfile, datasets, 'in_data')

    def __checkpoint_reload(self, nxsfile, datasets):
        cp = self.exp.checkpoint
        level, completed_plugins = cp._get_checkpoint_params()
        datasets = self._last_unique_datasets(datasets, completed_plugins+1)
        self._create_datasets(nxsfile, datasets, 'in_data')

        # update input data meta data
        for name in list(self.exp.index['in_data'].keys()):
            self.__update_metadata('in_data', name)

        if level == 'subplugin':
            # update output data meta data
            for name in list(self.exp.index['out_data'].keys()):
                self.__update_metadata('out_data', name)

    def __get_parameter_datasets(self, datasets):
        param_datasets = self.parameters['datasets']
        names = self.parameters['names'] if self.parameters['names'] else None
        if names and len(names) != len(param_datasets):
            raise Exception('Please enter a name for each dataset.')

        subset = {}
        found = []
        for i, p in enumerate(param_datasets):
            for d in datasets:
                if p in d['group'].name:
                    found.append(p)
                    name = names[i] if names else d['name']
                    subset[name] = d['group']

        missing = set(param_datasets).difference(set(found))
        if missing:
            msg = "Cannot find the dataset %s in the input nexus file." \
                % missing
            raise Exception(msg)

        if len(subset) != len(param_datasets):
            msg = "Multiple datasets with the same name cannot co-exist."
            raise Exception(msg)

        return subset

    def __update_metadata(self, dtype, name):
        cp = self.exp.checkpoint
        if name in cp.meta_data.get(dtype):
            new_dict = cp.meta_data.get([dtype, name])
            self.exp.index[dtype][name].meta_data._set_dictionary(new_dict)

    def _read_nexus_file(self, nxsfile, datasets):
        # find NXdata
        for key, value in nxsfile.items():
            if self._is_nxdata(value):
                datasets.append(self._get_dataset_info(key, value))
            elif isinstance(value, h5py.Group) and key != 'input_data':
                self._read_nexus_file(value, datasets)
        return datasets

    def _is_nxdata(self, value):
        check = 'NX_class' in value.attrs.keys() and ensure_string(value.attrs['NX_class']) == 'NXdata'
        return check

    def _get_dataset_info(self, key, value):
        import unicodedata
        key = unicodedata.normalize('NFKD', key)
        ksplit = key.split('-')

        if len(ksplit) == 1 and ''.join(key.split('_')[0:2]) == 'finalresult':
            name = '_'.join(key.split('_')[2:])
            pos = 'final'
        else:
            name = ''.join(ksplit[2:])
            pos = ksplit[0]
        return {'name': name, 'pos': pos, 'group': value}

    def _last_unique_datasets(self, datasets, final=None, names=None):
        if final:
            datasets = [d for d in datasets if int(d['pos']) < final]

        all_names = list(set([d['name'] for d in datasets]))
        names = [n for n in all_names if n in names] if names else all_names
        entries = {}
        for n in names:
            this_name = [d for d in datasets if d['name'] == n]
            max_pos = np.max(np.array([int(d['pos']) for d in this_name]))
            entries[n] = \
                [d['group'] for d in this_name if int(d['pos']) == max_pos][0]
        return entries

    def _create_datasets(self, nxsfile, datasets, dtype):
        data_objs = []

        for name, group in datasets.items():
            self.__set_preview_params(name)
            dObj = self._create_dataset(name, dtype)
            self._set_data_type(dObj, group, nxsfile.filename)
            self._read_nexus_group(group, dObj)
            dObj.set_shape(dObj.data.shape)
            self.__apply_previewing(dObj)
            data_objs.append(dObj)
        return data_objs

    def __set_preview_params(self, name):
        if isinstance(self._all_preview_params, dict):
            self.parameters['preview'] = self._all_preview_params[name] if \
                name in list(self._all_preview_params.keys()) else []

    def _set_data_type(self, dObj, group, nxs_filename):
        link = group.get(group.attrs['signal'], getlink=True)
        if isinstance(link, h5py._hl.group.HardLink) and \
                self.exp.meta_data.get('test_state') is True:
            link.filename = nxs_filename
            link.path = group.name + '/data'

        fname = os.path.join(os.path.dirname(nxs_filename), link.filename)

        dObj.backing_file = h5py.File(fname, 'r')
        dObj.data = dObj.backing_file[link.path]

        if 'data_type' not in group:
            return

        entry = group['data_type']
        args = self._get_data(entry, 'args')
        args = [args[''.join(['args', str(i)])] for i in range(len(args))]
        args = [a if a != 'self' else dObj for a in args]
        kwargs = self._get_data(entry, 'kwargs')
        extras = self._get_data(entry, 'extras')

        cls = str(self._get_data(entry, 'cls'))
        cls_split = cls.split('.')
        cls_inst = \
            pu.load_class('.'.join(cls_split[:-1]), cls_name=cls_split[-1])
        dObj.data = cls_inst(*args, **kwargs)
        dObj.data._base_post_clone_updates(dObj.data, extras)

    def _get_data(self, entry, key):
        if isinstance(entry[key], h5py.Group):
            ddict = {}
            for subkey in entry[key]:
                ddict[subkey] = self._get_data(entry[key], subkey)
            return ddict
        else:
            try:
                value = json.loads(entry[key][()][0])
            except Exception:
                value = cu._savu_decoder(entry[key][()])
            return value

    def _create_dataset(self, name, dtype):
        return self.exp.create_data_object(dtype, name, override=True)

    def _read_nexus_group(self, group, dObj):
        self._add_axis_labels(dObj, group)
        self._add_patterns(dObj, group)
        self._add_meta_data(dObj, group)

    def _add_axis_labels(self, dObj, group):
        axes = group.attrs['axes']
        ordered_axes = [None]*len(axes)
        axis_labels = []

        for ax in axes:
            ax = ensure_string(ax)
            ordered_axes[group.attrs['_'.join((ax, 'indices'))]] = ax
            dObj.meta_data.set(ax, group[ax][:])
            units = ensure_string(group[ax].attrs['units'])
            axis_labels.append('.'.join((ax, units)))

        dObj.set_axis_labels(*axis_labels)

    def _add_patterns(self, dObj, group):
        patterns = group['patterns']
        for key, value in patterns.items():
            dObj.add_pattern(key, core_dims=value['core_dims'],
                             slice_dims=value['slice_dims'])

    def _add_meta_data(self, dObj, group):
        def get_meta_data_entries(name, obj):
            for key, val in obj.attrs.items():
                if val == 'NXdata':
                    dObj.meta_data.set(name.split('/'), list(obj.values())[0][...])
        group['meta_data'].visititems(get_meta_data_entries)

    def _update_plugin_numbers(self, datasets):
        all_names = list(set([d['name'] for d in datasets]))
        updated = []
        for n in all_names:
            this = [d for d in datasets if d['name'] == n]
            p_numbers = [int(d['pos']) for d in this if d['pos'] != 'final']
            nPlugins = max(p_numbers)+1 if this and p_numbers else 1
            for d in this:
                if d['pos'] == 'final':
                    d['pos'] = nPlugins
            updated.extend(this)
        return datasets

    def __apply_previewing(self, dObj):
        preview = self._all_preview_params
        if isinstance(preview, dict):
            name = dObj.get_name()
            if name in list(self._all_preview_params.keys()):
                self.parameters['preview'] = self._all_preview_params[name]
        self.set_data_reduction_params(dObj)



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
.. module:: plugins_test
   :platform: Unix
   :synopsis: unittest test classes for plugins

.. moduleauthor:: Mark Basham <scientificsoftware@diamond.ac.uk>

"""

import unittest
import savu.test.test_utils as tu
from savu.test.travis.framework_tests.plugin_runner_test import \
        run_protected_plugin_runner

class FbpTest(unittest.TestCase):
    global data_file, experiment
    data_file = '24737.nxs'
    experiment = 'tomo'

    def test_fbp(self):
        process_list = 'reconstruction/simple_recon_test_process.nxs'
        options = tu.initialise_options(data_file, experiment, process_list)
        run_protected_plugin_runner(options)
        tu.cleanup(options)
                
#class ScikitimageSartTest(unittest.TestCase):
#
#    def test_scikit_sart(self):
#        options = tu.set_experiment('tomo')
#        plugin = 'savu.plugins.reconstructions.scikitimage_sart'
#        loader_dict = {'starts': [0, 0, 0],
#                       'stops': [-1, -1, -1],
#                       'steps': [20, 20, 20]}
#        data_dict = {'in_datasets': ['tomo'], 'out_datasets': ['test']}
#        saver_dict = {}
#        all_dicts = [loader_dict, data_dict, saver_dict]
#        run_protected_plugin_runner_no_process_list(options, plugin, all_dicts)

if __name__ == "__main__":
    unittest.main()



# Copyright 2019 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
.. module:: mipmaping plugin (a pyramid-like data downampling)
   :platform: Unix
   :synopsis:A plugin to downsample multidimensional data
.. moduleauthor:: Mark Basham & Daniil Kazantsev <scientificsoftware@diamond.ac.uk>
"""

import math
import h5py
import logging
import numpy as np
import skimage.measure as skim

from savu.plugins.plugin import Plugin
from savu.plugins.utils import register_plugin
from savu.plugins.driver.cpu_plugin import CpuPlugin


@register_plugin
class Mipmap(Plugin, CpuPlugin):
    """
    A plugin to downsample multidimensional data successively by powers of 2.\
    The output is multiple 'mipmapped' datasets, each a power of 2 smaller in \
    each dimension than the previous dataset.

    :u*param mode: One of 'mean', 'median', 'min', 'max'. Default: 'mean'.
    :u*param n_mipmaps:  The number of successive downsamples of powers of 2 \
        (e.g. n_mipmaps=3 implies downsamples (of the original data) of \
        binsize 1, 2 and 4 in each dimension).  Default: 3.
    :param out_dataset_prefix: The name of the dataset, to which the binsize \
        will be appended for each instance.  Default: 'Mipmap'.
    :*param out_datasets: Hidden out_datasets list as this is created \
        dynamically. Default: []
    """

    def __init__(self):
        super(Mipmap, self).__init__("Mipmap")

    def process_frames(self, data):
        self.mode_dict = { 'mean'  : np.mean,
                           'median': np.median,
                           'min'   : np.min,
                           'max'   : np.max }
        if self.parameters['mode'] in self.mode_dict:
            sampler = self.mode_dict[self.parameters['mode']]
        else:
            logging.warning("Unknown downsample mode. Using 'mean'.")
            sampler = np.mean

        inputMap = data[0]
        res = [data[0]]
        for i in range(1,self.parameters["n_mipmaps"]):
            downsample = skim.block_reduce(inputMap, (2, 2, 2), sampler)
            res.append(downsample)
            inputMap = np.copy(downsample)

        return res

    def setup(self):
        in_dataset, out_dataset = self.get_datasets()
        in_pData, out_pData = self.get_plugin_datasets()

        full_data_shape = list(in_dataset[0].get_shape())
        axis_labels = in_dataset[0].get_axis_labels()
        voxel_dims = \
            [i for i, e in enumerate(axis_labels) if 'voxel' in list(e.keys())[0]]

        # Sort out input data
        max_frames = self.get_max_frames()
        in_pData[0].plugin_data_setup('VOLUME_XZ', max_frames,
                slice_axis='voxel_y')

        # use this for 3D data (need to keep slice dimension)
        out_dataset = self.get_out_datasets()
        out_pData = self.get_plugin_out_datasets()

        for i in range(len(out_dataset)):
            shape = tuple([int(math.ceil(float(x)/(2**i))) if d in voxel_dims
                           else x for d, x in enumerate(full_data_shape)])
            out_dataset[i].create_dataset(axis_labels=in_dataset[0],
                                          patterns=in_dataset[0],
                                          shape=shape)
            out_pData[i].plugin_data_setup('VOLUME_XZ', max_frames // 2**i, slice_axis='voxel_y')

    def nInput_datasets(self):
        return 1

    def nOutput_datasets(self):
        n_mipmaps = self.parameters['n_mipmaps']
        name = self.parameters['out_dataset_prefix']
        self.parameters['out_datasets'] = \
            ['%s_%i' % (name, 2**i) for i in range(n_mipmaps)]
        return n_mipmaps

    def get_max_frames(self):
        return 8

    def fix_transport(self):
        return 'hdf5'



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: all_cpus_plugin
   :platform: Unix
   :synopsis: Base class for all plugins which want to use all CPU's on the
       target machine

.. moduleauthor:: Mark Basham <scientificsoftware@diamond.ac.uk>

"""
import logging


class AllCpusPlugin(object):
    """
    The base class from which all plugins should inherit.
    """

    def __init__(self):
        super(AllCpusPlugin, self).__init__()
        self.available_cpus = 1

    def run_process(self, data, output, processes, process):
        count = 0
        cpu_processes = []
        self.available_cpus = sum(["CPU" in i for i in processes])
        for i in ["CPU0" in i for i in processes]:
            if i:
                cpu_processes.append(count)
                count += 1
            else:
                cpu_processes.append(-1)
        if cpu_processes[process] >= 0:
            logging.debug("Running the CPUS Process %i", process)
            new_processes = [i for i in processes if "CPU0" in i]
            #logging.debug(new_processes)
            #logging.debug(cpu_processes)
            logging.debug("Process is %s",
                          new_processes[cpu_processes[process]])
            self.pre_process(0 if isinstance(data, str) else data.get_data_shape())
            self.process(data, output, new_processes,
                                cpu_processes[process])
            self.post_process()
            return
        logging.debug("Not Running the task as not CPU0")
        return

    def process(self, data, output, processes, process):
        """
        This method is called after the plugin has been created by the
        pipeline framework, self.available_cpus contains the number of
        cpus which can be used by this process

        :param data: The input data object.
        :type data: savu.data.data_structures
        :param data: The output data object
        :type data: savu.data.data_structures
        :param processes: The number of processes which will be doing the work
        :type path: int
        :param path: The specific process which we are
        :type path: int
        """
        logging.error("process needs to be implemented for proc %i of %i :" +
                      " input is %s and output is %s",
                      process, processes, data.__class__, output.__class__)
        raise NotImplementedError("process needs to be implemented")



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: plugin_list
   :platform: Unix
   :synopsis: Contains the PluginList class, which deals with loading and \
   saving the plugin list, and the CitationInformation class. An instance is \
   held by the MetaData class.

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""
import ast
import copy
import inspect
import json
import logging
import os
import re
from collections import defaultdict

import h5py
import numpy as np
import yaml

import savu.data.framework_citations as fc
import savu.plugins.loaders.utils.yaml_utils as yu
import savu.plugins.utils as pu
from savu.data.meta_data import MetaData

NX_CLASS = 'NX_class'


class PluginList(object):
    """
    The PluginList class handles the plugin list - loading, saving and adding
    citation information for the plugin
    """

    def __init__(self):
        self.plugin_list = []
        self.n_plugins = None
        self.n_loaders = 0
        self.n_savers = 0
        self.loader_idx = None
        self.saver_idx = None
        self.datasets_list = []
        self.saver_plugin_status = True
        self._template = None
        self.version = None

    def add_template(self, create=False):
        self._template = Template(self)
        if create:
            self._template.creating = True

    def _get_plugin_entry_template(self):
        template = {'active': True,
                    'name': None,
                    'id': None,
                    'desc': None,
                    'data': None,
                    'user': [],
                    'hide': []}
        return template

    def __get_json_keys(self):
        return ['data', 'desc', 'user', 'hide']

    def _populate_plugin_list(self, filename, active_pass=False, template=False):
        """ Populate the plugin list from a nexus file. """
        plugin_file = h5py.File(filename, 'r')

        if 'entry/savu_notes/version' in plugin_file:
            self.version = plugin_file['entry/savu_notes/version'][()]

        plugin_group = plugin_file['entry/plugin']
        self.plugin_list = []
        single_val = ['name', 'id', 'pos', 'active']
        exclude = ['citation']
        for group in plugin_group.keys():
            plugin = self._get_plugin_entry_template()
            entry_keys = plugin_group[group].keys()
            parameters = [k for k in entry_keys for e in exclude if k not in
                          single_val and e not in k]

            if 'active' in entry_keys:
                plugin['active'] = plugin_group[group]['active'][0]

            if plugin['active'] or active_pass:
                plugin['name'] = plugin_group[group]['name'][0].decode("utf-8")
                plugin['id'] = plugin_group[group]['id'][0].decode("utf-8")
                plugin['pos'] = group.strip()

                for param in parameters:
                    try:
                        plugin[param] = json.loads(plugin_group[group][param][0])
                    except ValueError as e:
                        raise ValueError(f"Error: {e}\n Could not parse key '{param}' from group '{group}' as JSON")
                self.plugin_list.append(plugin)

        if template:
            self.add_template()
            self._template.update_process_list(template)

        plugin_file.close()

    def _save_plugin_list(self, out_filename):
        with h5py.File(out_filename, 'a') as nxs_file:

            entry = nxs_file.require_group('entry')

            self._save_framework_citations(self._overwrite_group(
                entry, 'framework_citations', 'NXcollection'))

            self.__save_savu_notes(self._overwrite_group(
                entry, 'savu_notes', 'NXnote'))

            plugins_group = self._overwrite_group(entry, 'plugin', 'NXprocess')

            count = 1
            for plugin in self.plugin_list:
                self.__populate_plugins_group(plugins_group, plugin, count)

        if self._template and self._template.creating:
            fname = os.path.splitext(out_filename)[0] + '.savu'
            self._template._output_template(fname, out_filename)

    def _overwrite_group(self, entry, name, nxclass):
        if name in entry:
            entry.pop(name)
        group = entry.create_group(name.encode("ascii"))
        group.attrs[NX_CLASS] = nxclass.encode("ascii")
        return group

    def __save_savu_notes(self, notes):
        from savu.version import __version__
        notes['version'] = __version__

    def __populate_plugins_group(self, plugins_group, plugin, count):
        if 'pos' in plugin.keys():
            num = int(re.findall(r'\d+', plugin['pos'])[0])
            letter = re.findall('[a-z]', plugin['pos'])
            letter = letter[0] if letter else ""
            group_name = "%*i%*s" % (4, num, 1, letter)
        else:
            group_name = "%*i" % (4, count)

        plugin_group = plugins_group.create_group(group_name.encode("ascii"))

        plugin_group.attrs[NX_CLASS] = 'NXnote'.encode('ascii')
        required_keys = self._get_plugin_entry_template().keys()
        json_keys = self.__get_json_keys()

        if 'cite' in plugin.keys():
            if plugin['cite'] is not None:
                self._output_plugin_citations(plugin['cite'], plugin_group)

        for key in required_keys:
            # only need to apply dumps if saving in configurator
            data = self.__dumps(plugin[key]) if key == 'data' else plugin[key]

            # get the string value
            data = json.dumps(data) if key in json_keys else plugin[key]
            # if the data is string it has to be encoded to ascii so that
            # hdf5 can save out the bytes
            if isinstance(data, str):
                data = data.encode("ascii")
            data = np.array([data])
            plugin_group.create_dataset(key.encode('ascii'), data.shape, data.dtype, data)

    def __dumps(self, data_dict):
        """ Replace any missing quotes around variables
        """
        for key, val in data_dict.items():
            if isinstance(val, str):
                try:
                    data_dict[key] = ast.literal_eval(val)
                    continue
                except Exception:
                    pass
                try:
                    data_dict[key] = yaml.load(val, Loader=yaml.SafeLoader)
                    continue
                except Exception:
                    pass
                try:
                    isdict = re.findall(r"[\{\}]+", val)
                    if isdict:
                        val = val.replace("[", "'[").replace("]", "]'")
                        data_dict[key] = self.__dumps(yaml.load(val))
                    else:
                        data_dict[key] = pu.parse_config_string(val)
                    continue
                except Exception:
                    # for when parameter tuning with lists is added to the framework
                    if len(val.split(';')) > 1:
                        pass
                    else:
                        raise Exception("Invalid string %s" % val)
        return data_dict

    def _add(self, idx, entry):
        self.plugin_list.insert(idx, entry)
        self.__set_loaders_and_savers()

    def _remove(self, idx):
        del self.plugin_list[idx]
        self.__set_loaders_and_savers()

    def _output_plugin_citations(self, citations, group):
        if not isinstance(citations, list):
            citations = [citations]
        for cite in citations:
            citation_group = group.create_group(cite.name.encode("ascii"))
            cite.write(citation_group)

    def _save_framework_citations(self, group):
        framework_cites = fc.get_framework_citations()
        for cite in framework_cites:
            citation_group = group.require_group(cite['name'].encode("ascii"))
            citation = CitationInformation()
            citation.name = cite["name"]
            citation.description = cite["description"]
            citation.bibtex = cite["bibtex"]
            citation.endnote = cite["endnote"]

            citation.write(citation_group)

    def _get_docstring_info(self, plugin):
        plugin_inst = pu.plugins[plugin]()
        plugin_inst._populate_default_parameters()
        return plugin_inst.docstring_info

    # def _byteify(self, input):
    #     if isinstance(input, dict):
    #         return {self._byteify(key): self._byteify(value)
    #                 for key, value in input.items()}
    #     elif isinstance(input, list):
    #         temp = [self._byteify(element) for element in input]
    #         return temp
    #     elif isinstance(input, str):
    #         return input.encode('utf-8')
    #     else:
    #         return input

    def _set_datasets_list(self, plugin):
        in_pData, out_pData = plugin.get_plugin_datasets()
        in_data_list = self._populate_datasets_list(in_pData)
        out_data_list = self._populate_datasets_list(out_pData)
        self.datasets_list.append({'in_datasets': in_data_list,
                                   'out_datasets': out_data_list})

    def _populate_datasets_list(self, data):
        data_list = []
        for d in data:
            name = d.data_obj.get_name()
            pattern = copy.deepcopy(d.get_pattern())
            pattern[list(pattern.keys())[0]]['max_frames_transfer'] = \
                d.meta_data.get('max_frames_transfer')
            pattern[list(pattern.keys())[0]]['transfer_shape'] = \
                d.meta_data.get('transfer_shape')
            data_list.append({'name': name, 'pattern': pattern})
        return data_list

    def _get_datasets_list(self):
        return self.datasets_list

    def _reset_datasets_list(self):
        self.datasets_list = []

    def _get_n_loaders(self):
        return self.n_loaders

    def _get_n_savers(self):
        return self.n_savers

    def _get_loaders_index(self):
        return self.loader_idx

    def _get_savers_index(self):
        return self.saver_idx

    def _get_n_processing_plugins(self):
        return len(self.plugin_list) - self._get_n_loaders()

    def __set_loaders_and_savers(self):
        """ Get lists of loader and saver positions within the plugin list and
        set the number of loaders.

        :returns: loader index list and saver index list
        :rtype: list(int(loader)), list(int(saver))
        """
        from savu.plugins.loaders.base_loader import BaseLoader
        from savu.plugins.savers.base_saver import BaseSaver
        loader_idx = []
        saver_idx = []
        self.n_plugins = len(self.plugin_list)

        for i in range(self.n_plugins):
            pid = self.plugin_list[i]['id']
            bases = inspect.getmro(pu.load_class(pid))
            loader_list = [b for b in bases if b == BaseLoader]
            saver_list = [b for b in bases if b == BaseSaver]
            if loader_list:
                loader_idx.append(i)
            if saver_list:
                saver_idx.append(i)
        self.loader_idx = loader_idx
        self.saver_idx = saver_idx
        self.n_loaders = len(loader_idx)
        self.n_savers = len(saver_idx)

    def _check_loaders(self):
        """ Check plugin list starts with a loader and ends with a saver.
        """
        self.__set_loaders_and_savers()
        loaders = self._get_loaders_index()

        if loaders:
            if loaders[0] != 0 or loaders[-1] + 1 != len(loaders):
                raise Exception("All loader plugins must be at the beginning "
                                "of the plugin list")
        else:
            raise Exception("The first plugin in the plugin list must be a "
                            "loader plugin.")

    def _add_missing_savers(self, exp):
        """ Add savers for missing datasets. """
        data_names = exp.index['in_data'].keys()
        saved_data = []
        
        for i in self._get_savers_index():
            saved_data.append(self.plugin_list[i]['data']['in_datasets'])
        saved_data = set([s for sub_list in saved_data for s in sub_list])

        for name in [data for data in data_names if data not in saved_data]:
            pos = exp.meta_data.get('nPlugin')+1
            exp.meta_data.set('nPlugin', pos)
            process = {}
            plugin = pu.load_class('savu.plugins.savers.hdf5_saver')()
            plugin.parameters['in_datasets'] = [name]
            process['name'] = plugin.name
            process['id'] = plugin.__module__
            process['pos'] = str(pos+1)
            process['data'] = plugin.parameters
            process['active'] = True
            process['desc'] = plugin.parameters_desc
            self._add(pos+1, process)

    def _update_datasets(self, plugin_no, data_dict):
        n_loaders = self._get_n_loaders()
        idx = self._get_n_loaders() + plugin_no
        self.plugin_list[idx]['data'].update(data_dict)

    def _get_dataset_flow(self):
        datasets_idx = []
        n_loaders = self._get_n_loaders()
        n_plugins = self._get_n_processing_plugins()
        for i in range(self.n_loaders, n_loaders + n_plugins):
            datasets_idx.append(self.plugin_list[i]['data']['out_datasets'])
        return datasets_idx

    def _contains_gpu_processes(self):
        """ Returns True if gpu processes exist in the process list. """
        try:
            from savu.plugins.driver.gpu_plugin import GpuPlugin
            for i in range(self.n_plugins):
                bases = inspect.getmro(pu.load_class(self.plugin_list[i]['id']))
                if GpuPlugin in bases:
                    return True
        except ImportError as ex:
            if "pynvml" in ex.message:
                logging.error('Error while importing GPU dependencies: %s',
                              ex.message)
            else:
                raise

        return False


class Template(object):
    """ A class to read and write templates for plugin lists.
    """

    def __init__(self, plist):
        super(Template, self).__init__()
        self.plist = plist
        self.creating = False

    def _output_template(self, fname, process_fname):
        plist = self.plist.plugin_list
        index = [i for i in range(len(plist)) if plist[i]['active']]

        local_dict = MetaData(ordered=True)
        global_dict = MetaData(ordered=True)
        local_dict.set(['process_list'], os.path.abspath(process_fname))

        for i in index:
            params = self.__get_template_params(plist[i]['data'], [])
            name = plist[i]['name']
            for p in params:
                ptype, isyaml, key, value = p
                if isyaml:
                    data_name = isyaml if ptype == 'local' else 'all'
                    local_dict.set([i + 1, name, data_name, key], value)
                elif ptype == 'local':
                    local_dict.set([i + 1, name, key], value)
                else:
                    global_dict.set(['all', name, key], value)

        with open(fname, 'w') as stream:
            local_dict.get_dictionary().update(global_dict.get_dictionary())
            yu.dump_yaml(local_dict.get_dictionary(), stream)

    def __get_template_params(self, params, tlist, yaml=False):
        for key, value in params.items():
            if key == 'yaml_file':
                yaml_dict = self._get_yaml_dict(value)
                for entry in list(yaml_dict.keys()):
                    self.__get_template_params(
                        yaml_dict[entry]['params'], tlist, yaml=entry)
            value = pu.is_template_param(value)
            if value is not False:
                ptype, value = value
                isyaml = yaml if yaml else False
                tlist.append([ptype, isyaml, key, value])
        return tlist

    def _get_yaml_dict(self, yfile):
        from savu.plugins.loaders.yaml_converter import YamlConverter
        yaml = YamlConverter()
        template_check = pu.is_template_param(yfile)
        yfile = template_check[1] if template_check is not False else yfile
        yaml.parameters = {'yaml_file': yfile}
        return yaml.setup(template=True)

    def update_process_list(self, template):
        tdict = yu.read_yaml(template)
        del tdict['process_list']

        for plugin_no, entry in tdict.items():
            plugin = list(entry.keys())[0]
            for key, value in list(entry.values())[0].iteritems():
                depth = self.dict_depth(value)
                if depth == 1:
                    self._set_param_for_template_loader_plugin(
                        plugin_no, key, value)
                elif depth == 0:
                    if plugin_no == 'all':
                        self._set_param_for_all_instances_of_a_plugin(
                            plugin, key, value)
                    else:
                        data = self._get_plugin_data_dict(str(plugin_no))
                        data[key] = value
                else:
                    raise Exception("Template key not recognised.")

    def dict_depth(self, d, depth=0):
        if not isinstance(d, dict) or not d:
            return depth
        return max(self.dict_depth(v, depth + 1) for k, v in d.items())

    def _set_param_for_all_instances_of_a_plugin(self, plugin, param, value):
        # find all plugins with this name and replace the param
        for p in self.plist.plugin_list:
            if p['name'] == plugin:
                p['data'][param] = value

    def _set_param_for_template_loader_plugin(self, plugin_no, data, value):
        param_key = list(value.keys())[0]
        param_val = list(value.values())[0]
        pdict = self._get_plugin_data_dict(str(plugin_no))['template_param']
        pdict = defaultdict(dict) if not pdict else pdict
        pdict[data][param_key] = param_val

    def _get_plugin_data_dict(self, plugin_no):
        """ input plugin_no as a string """
        plist = self.plist.plugin_list
        index = [plist[i]['pos'] for i in range(len(plist))]
        return plist[index.index(plugin_no)]['data']


class CitationInformation(object):
    """
    Descriptor of Citation Information for plugins
    """

    name: str = 'citation'
    bibtex: str = "Default Bibtex"
    description: str = "Default Description"
    doi: str = "Default DOI"
    endnote: str = "Default Endnote"

    def write(self, citation_group):
        # classes don't have to be encoded to ASCII
        citation_group.attrs[NX_CLASS] = 'NXcite'
        description_array = np.array([self.description.encode('ascii')])
        citation_group.create_dataset('description'.encode('ascii'),
                                      description_array.shape,
                                      description_array.dtype,
                                      description_array)
        doi_array = np.array([self.doi.encode('ascii')])
        citation_group.create_dataset('doi'.encode('ascii'),
                                      doi_array.shape,
                                      doi_array.dtype,
                                      doi_array)
        endnote_array = np.array([self.endnote.encode('ascii')])
        citation_group.create_dataset('endnote'.encode('ascii'),
                                      endnote_array.shape,
                                      endnote_array.dtype,
                                      endnote_array)
        bibtex_array = np.array([self.bibtex.encode('ascii')])
        citation_group.create_dataset('bibtex'.encode('ascii'),
                                      bibtex_array.shape,
                                      bibtex_array.dtype,
                                      bibtex_array)



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: base_absorption_correction
   :platform: Unix
   :synopsis: A set of plugins to correct for absorption effects in xrd and\
       xrf ct.

.. moduleauthor:: Aaron D. Parsons <scientificsoftware@diamond.ac.uk>
"""

import logging
from savu.plugins.plugin import Plugin
from savu.plugins.driver.cpu_plugin import CpuPlugin
import xraylib as xl


class BaseAbsorptionCorrection(Plugin, CpuPlugin):
    """
    a base absorption correction for stxm and xrd
    :param azimuthal_offset: angle between detectors. Default: -90.0.
    :param density: the density. Default: 3.5377.
    :param compound: the compount. Default: 'Co0.1Re0.01Ti0.05(SiO2)0.84'.
    """

    def __init__(self, name):
        logging.debug("Starting Absorption correction")
        super(BaseAbsorptionCorrection,
              self).__init__(name)

    def setup(self):
        logging.debug('setting up the base absorption correction')
        in_dataset, out_datasets = self.get_datasets()
        in_pData, out_pData = self.get_plugin_datasets()
        in_meta_data = in_dataset[0].meta_data

        try:
            idx = in_meta_data.get('PeakEnergy')
        except KeyError:
            logging.debug('No PeakEnergy: trying with the fullSpectrum')
            try:
                in_dataset, out_datasets = self.get_datasets()
                in_dataset[0].get_data_patterns()['SPECTRUM']
                idx = list(in_meta_data.get('energy'))
            except KeyError:
                logging.debug("No PeakEnergy or energy axis. This won't work")
                raise
        self.nChannels = len(idx)
        in_pData[0].plugin_data_setup('SINOGRAM', self.get_num_channels())
        in_pData[1].plugin_data_setup('SINOGRAM', 'single')
        spectra = out_datasets[0]
        spectra.create_dataset(in_dataset[0])

        out_pData[0].plugin_data_setup('SINOGRAM', self.get_num_channels())

    def get_num_channels(self):
        return self.nChannels

    def get_azimuthal_offset(self):
        return self.parameters['azimuthal_offset']

    def nInput_datasets(self):
        return 2

    def nOutput_datasets(self):
        return 1

    def get_mu(self, compound, energy, density):
        '''
        returns mu for a compound for a single, or list, of energies
        '''
        if isinstance(energy, (list)):
            op = []
            for e in energy:
                op.append((xl.CS_Total_CP(compound, e)*density)*1e-2)
            return op
        else:
            return (xl.CS_Total_CP(compound, energy)*density)*1e-2



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: I14FluoLoader
   :platform: Unix
   :synopsis: A class for loading I14s xrf data

.. moduleauthor:: Aaron Parsons <scientificsoftware@diamond.ac.uk>

"""

from savu.plugins.loaders.base_loader import BaseLoader
import numpy as np
from savu.plugins.utils import register_plugin
import logging
import h5py as h5

@register_plugin
class I14FluoLoader(BaseLoader):
    """
    A class to load i14s xrf data
    :param mono_path: The mono energy. Default: '/entry/instrument/beamline/DCM/dcm_energy'.
    """

    def __init__(self, name='I14FluoLoader'):
        super(I14FluoLoader, self).__init__(name)

    def setup(self):
        """
         Define the input nexus file

        :param path: The full path of the NeXus file to load.
        :type path: str
        """

        det_path = 'entry/xsp3_addetector/'
        exp = self.exp
        data_obj = exp.create_data_object("in_data", 'fluo')
        data_obj.backing_file = \
            h5.File(exp.meta_data.get("data_file"), 'r')
        data_obj.data = data_obj.backing_file[det_path+'/data']
        data_obj.set_shape(data_obj.data.shape)
        entry = h5.File(exp.meta_data.get("data_file"),'r')[det_path]
        scan_axis = entry.attrs['axes']
        scan_axis = [str(ix) for ix in scan_axis]
        if list(scan_axis).count('.')>1:
            logging.debug("Detector channels not summed.")

            for axis in scan_axis:
                if axis is not '.':
                    axis_data = entry[str(axis)][...]
                    data_obj.meta_data.set(axis,axis_data)
            scan_axis[-2] = 'idx'
            scan_axis[-1] = 'energy'
            scan_axis = [ix +'.units' for ix in scan_axis]

        else:
            logging.debug("Detector channels summed.")
            for axis in scan_axis:
                if axis is not '.':
                    axis_data = entry[str(axis)][...]

                    data_obj.meta_data.set(axis,axis_data)
            scan_axis[-1] = 'energy'
            scan_axis = [ix +'.units' for ix in scan_axis]
        try:
            mono_energy = h5.File(exp.meta_data.get("data_file"))[self.parameters['mono_path']][...]

            data_obj.meta_data.set('mono_energy',mono_energy)
        except KeyError:
            logging.warning("Monochromator value not found at %s, are you sure this is correct? Continuing anyway..." % self.parameters['mono_path'])
        # axis label
#         print "the labels are:"+str(labels)
        data_obj.set_axis_labels(*tuple(scan_axis))

        dims = list(range(len(data_obj.get_shape())))
        spec_core = (-1,) # it will always be this
#         print spec_core

        spec_slice = tuple(dims[:-1])

        logging.debug("is a spectrum")
        logging.debug("the spectrum cores are:"+str(spec_core))
        logging.debug("the spectrum slices are:"+str(spec_slice))
        data_obj.add_pattern("SPECTRUM", core_dims=spec_core,
                             slice_dims=spec_slice)


        positions_label = (data_obj.get_data_dimension_by_axis_label('x', contains=True),data_obj.get_data_dimension_by_axis_label('y', contains=True))

        data_obj.add_pattern("PROJECTION", core_dims=positions_label, slice_dims=tuple(set(dims)-set(positions_label)))

        self.set_data_reduction_params(data_obj)



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: multi_threaded_plugin
   :platform: Unix
   :synopsis: Driver to run one instance of a plugin on EACH node.

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""
from mpi4py import MPI

from savu.plugins.driver.plugin_driver import PluginDriver


class MultiThreadedPlugin(PluginDriver):
    """
    Initiates a multi-theaded plugin on one core of each node.

    """

    def __init__(self):
        super(MultiThreadedPlugin, self).__init__()

    def get_mem_multiply(self):
        self.processes = self.exp.meta_data.get("processes")
        self.nNodes = self.processes.count(self.processes[0])
        return len(self.processes)/self.nNodes

    def _run_plugin(self, exp, transport):
        process = exp.meta_data.get("process")
        nCores = len(self.processes) // self.nNodes
        masters = self._get_masters(self.processes)

        self.__create_new_communicator(masters, exp)
        self.exp._barrier()

        if process in masters:
            self.parameters['available_CPUs'] = nCores
            self.parameters['available_GPUs'] = len([p for p in self.processes if 'GPU' in p]) // self.nNodes
            self._run_plugin_instances(transport, communicator=self.new_comm)
            self.__free_communicator()

        self.exp._barrier()
        return

    def _get_masters(self, processes):
        masters = [p for p in range(len(processes)) if processes[p] == 'GPU0']
        if not masters:
            masters = \
                [p for p in range(len(processes)) if processes[p] == 'CPU0']
        return masters

    def __create_new_communicator(self, ranks, exp):
        self.group = MPI.COMM_WORLD.Get_group()
        self.new_group = MPI.Group.Incl(self.group, ranks)
        self.new_comm = MPI.COMM_WORLD.Create(self.new_group)
        self.exp._barrier()

    def __free_communicator(self):
        self.group.Free()
        self.new_group.Free()
        self.new_comm.Free()

    def min_max_cpus(self):
        """ Sets the bounds on the number of CPUs required by the plugin, such
        that if bounds=[b1, b2] then b1 is the lower bound and b2 is the upper
        bound.  Set each entry to None if there are no bounds.
        """
        return [None, None]

    def min_max_gpus(self):
        """ Sets the bounds on the number of GPUs required by the plugin, such
        that if bounds=[b1, b2] then b1 is the lower bound and b2 is the upper
        bound.  Set each entry to None if there are no bounds and to 0 if no
        GPUs are required.
        """
        return [0, 0]



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: base_azimuthal_integrator
   :platform: Unix
   :synopsis: A plugin to integrate azimuthally "symmetric" signals i.e. \
       SAXS, WAXS or XRD.Requires a calibration file

.. moduleauthor:: Aaron D. Parsons <scientificsoftware@diamond.ac.uk>
"""

import copy
import pyFAI
import logging

import numpy as np
from savu.plugins.plugin import Plugin
from savu.plugins.driver.cpu_plugin import CpuPlugin


class BaseAzimuthalIntegrator(Plugin, CpuPlugin):
    """
    a base azimuthal integrator for pyfai

    :param use_mask: Should we mask. Default: False.
    :param num_bins: number of bins. Default: 1005.

    """

    def __init__(self, name='BaseAzimuthalIntegrator'):
        logging.debug("Starting 1D azimuthal integrationr")
        super(BaseAzimuthalIntegrator, self).__init__(name)

    def pre_process(self):
        """
        This method is called after the plugin has been created by the
        pipeline framework as a pre-processing step

        :param parameters: A dictionary of the parameters for this plugin, or
            None if no customisation is required
        :type parameters: dict
        """

        in_dataset, out_datasets = self.get_datasets()
        mData = self.get_in_meta_data()[0]
        in_d1 = in_dataset[0]

        ai = pyFAI.AzimuthalIntegrator()  # get me an integrator object

        # prep the goemtry
        px_m = mData.get('x_pixel_size')
        bc_m = [mData.get("beam_center_x"),
                mData.get("beam_center_y")]  # in metres
        bc = bc_m / px_m  # convert to pixels
        px = px_m*1e6  # convert to microns
        distance = mData.get('distance')*1e3  # convert to mm
        wl = mData.get('incident_wavelength')[...]  # in m
        self.wl = wl

        yaw = -mData.get("yaw")
        roll = mData.get("roll")

        ai.setFit2D(distance, bc[0], bc[1], yaw, roll, px, px, None)

        ai.set_wavelength(wl)
        logging.debug(ai)

        sh = in_d1.get_shape()

        if (self.parameters["use_mask"]):
            mask = mData.get("mask")
        else:
            mask = np.zeros((sh[-2], sh[-1]))
        # now integrate in radius (1D)print "hello"
        self.npts = self.get_parameters('num_bins')
        self.params = [mask, self.npts, mData, ai]
        # now set the axis values, we shouldn't do this in every slice

        axis, __remapped = \
            ai.integrate1d(data=mask, npt=self.npts, unit='q_A^-1',
                           correctSolidAngle=False)

        self.add_axes_to_meta_data(axis, mData)

    def setup(self):
        in_dataset, out_dataset = self.get_datasets()

        # AMEND THE PATTERNS: The output dataset will have one dimension less
        # than the in_dataset, so remove the final slice dimension from any
        # patterns you want to keep.
        rm_dim = str(in_dataset[0].get_data_patterns()
                     ['SINOGRAM']['slice_dims'][-1])
        patterns = ['SINOGRAM.' + rm_dim, 'PROJECTION.' + rm_dim]

        # AMEND THE AXIS LABELS: Find the dimensions to remove using their
        # axis_labels to ensure the plugin is as generic as possible and will
        # work for data in all orientations.
        axis_labels = copy.copy(in_dataset[0].get_axis_labels())
        rm_labels = ['detector_x', 'detector_y']
        rm_dims = sorted([in_dataset[0].get_data_dimension_by_axis_label(a)
                          for a in rm_labels])[::-1]
        for d in rm_dims:
            del axis_labels[d]
        # Add a new axis label to the list
        axis_labels.append({'Q': 'Angstrom^-1'})

        # AMEND THE SHAPE: Remove the two unrequired dimensions from the
        # original shape and add a new dimension shape.
        shape = list(in_dataset[0].get_shape())
        for d in rm_dims:
            del shape[d]
        shape += (self.get_parameters('num_bins'),)

        # populate the output dataset
        out_dataset[0].create_dataset(
                patterns={in_dataset[0]: patterns},
                axis_labels=axis_labels,
                shape=tuple(shape))

        spectrum = \
            {'core_dims': (-1,), 'slice_dims': tuple(range(len(shape)-1))}
        out_dataset[0].add_pattern("SPECTRUM", **spectrum)
        # =====================================================================

        # ================== populate plugin datasets =========================
        in_pData, out_pData = self.get_plugin_datasets()
        in_pData[0].plugin_data_setup('DIFFRACTION', 'single')
        out_pData[0].plugin_data_setup('SPECTRUM', 'single')
        # =====================================================================

    def get_max_frames(self):
        return 'single'

    def nOutput_datasets(self):
        return 1

    def add_axes_to_meta_data(self, axis, mData):
        qanstrom = axis
        dspacing = 2*np.pi/qanstrom
        ttheta = 2*180*np.arcsin(self.wl/(2*dspacing*1e-10))/np.pi
        mData.set('Q', qanstrom)
        mData.set('D', dspacing)
        mData.set('2Theta', ttheta)



# Copyright 2018 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: visual_hulls_recon
   :platform: Unix
   :synopsis: simple visual hulls reconstruction

.. moduleauthor:: Mark Basham <scientificsoftware@diamond.ac.uk>

"""

import logging
import numpy as np

from scipy.ndimage.filters import median_filter
from savu.plugins.reconstructions.base_recon import BaseRecon
from savu.data.plugin_list import CitationInformation
from savu.plugins.driver.cpu_plugin import CpuPlugin
from savu.plugins.utils import register_plugin


@register_plugin
class VisualHullsRecon(BaseRecon, CpuPlugin):
    """
    A Plugin to reconstruct an image by filter back projection
    using the inverse radon transform from scikit-image.

    :param threshold: threshold to binarize the input sinogram. Default: 0.5.

    :~param outer_pad: Not required. Default: False.
    :~param centre_pad: Not required. Default: False.
    """

    def __init__(self):
        logging.debug("initialising Scikitimage Filter Back Projection")
        logging.debug("Calling super to make sure that all superclasses are " +
                      " initialised")
        super(VisualHullsRecon,
              self).__init__("VisualHullsRecon")

    def _mapping_array(self, array_shape, center, theta):
        x, y = np.meshgrid(np.arange(-center, array_shape[0] - center),
                           np.arange(-center, array_shape[1] - center))
        return x*np.cos(theta) - y*np.sin(theta)

    def _recon_hull(self, sino, centre, angles):
        data_shape = (sino.shape[1], sino.shape[1])
        full = np.ones(data_shape)
        for i in range(len(angles)):
            mapping_array = self._mapping_array(data_shape, centre, np.deg2rad(angles[i]))
            mapping_array = np.clip(mapping_array.astype('int')+centre, 0,
                                    sino.shape[1]-1).astype('int')
            mask = sino[i, :][mapping_array]
            full -= 1-mask
        data_range = full.max() - full.min()
        full += data_range // 4
        full[full < 0.5] = 0
        return full

    def _binarize_sinogram(self, input_sinogram, threshold):
        sino = np.zeros_like(input_sinogram)
        sino[input_sinogram > threshold] = 1
        # as this is a simple routine, do a quick median filter to
        # get rid of any stray pixels in the binarization.
        sino = median_filter(sino, size=3)
        return sino

    def process_frames(self, data):
        sinogram = self._binarize_sinogram(data[0],
                                           self.parameters['threshold'])
        centre_of_rotations, angles, vol_shape, init = \
            self.get_frame_params()
        recon = self._recon_hull(sinogram, centre_of_rotations, angles)
        return recon

    def get_max_frames(self):
        return 'single'

    def get_citation_information(self):
        cite_info = CitationInformation()
        cite_info.description = \
            ("The reconstruction performed in this processing " +
             "chain is derived from this work.")
        cite_info.bibtex = \
            ("@article{laurentini1994visual," +
             "title={The visual hull concept for silhouette-based image understanding}," +
             "author={Laurentini, Aldo}," +
             "journal={IEEE Transactions on pattern analysis and machine intelligence}," +
             "volume={16}," +
             "number={2}," +
             "pages={150--162}," +
             "year={1994}," +
             "publisher={IEEE}" +
             "}")
        cite_info.endnote = \
            ("%0 Journal Article" +
             "%T The visual hull concept for silhouette-based image understanding" +
             "%A Laurentini, Aldo" +
             "%J IEEE Transactions on pattern analysis and machine intelligence" +
             "%V 16" +
             "%N 2" +
             "%P 150-162" +
             "%@ 0162-8828" +
             "%D 1994" +
             "%I IEEE")
        cite_info.doi = "http://dx.doi.org/10.1109/34.273735"
        return cite_info



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: tiff_saver
   :platform: Unix
   :synopsis: A class to save output in tiff format

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""

import tifffile as tf

from savu.plugins.savers.base_image_saver import BaseImageSaver
from savu.plugins.utils import register_plugin
from savu.plugins.driver.cpu_plugin import CpuPlugin


@register_plugin
class TiffSaver(BaseImageSaver, CpuPlugin):
    """
    A class to save tomography data to tiff files
    :param pattern: How to slice the data. Default: 'VOLUME_XZ'.
    :param prefix: Override the default output tiff file prefix. Default: None.

    :config_warn: Do not use this plugin if the raw data is greater than \
    100 GB.
    """

    def __init__(self, name='TiffSaver'):
        super(TiffSaver, self).__init__(name)

    def process_frames(self, data):
        frame = self.get_global_frame_index()[self.count]
        filename = '%s%05i.tiff' % (self.filename, frame)
        tf.imsave(filename, data[0])
        self.count += 1



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: dosna_transport_data
   :platform: Unix
   :synopsis: A data transport class that is inherited by Data class at \
   runtime. It organises the slice list and moves the data.

.. moduleauthor:: Emilio Perez Juarez <scientificsoftware@diamond.ac.uk>

"""

from savu.data.transport_data.hdf5_transport_data import Hdf5TransportData


class DosnaTransportData(Hdf5TransportData):
    """
    The DosnaTransportData class performs the organising and movement of data.
    """

    def __init__(self, data_obj, name='DosnaTransportData'):
        super(DosnaTransportData, self).__init__(data_obj)



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
.. module:: testing_iterative_plugin2
   :platform: Unix
   :synopsis: Iterative plugin example
.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>
"""

import numpy as np

from savu.plugins.utils import register_plugin
from savu.plugins.filters.base_filter import BaseFilter
from savu.plugins.driver.iterative_plugin import IterativePlugin


@register_plugin
class TestingIterativePlugin2(BaseFilter, IterativePlugin):
    """
    A plugin to test the iterative plugin driver
    """

    def __init__(self):
        super(TestingIterativePlugin2, self).\
            __init__("TestingIterativePlugin2")

    def pre_process(self):
        self.set_iterations(3)

    def process_frames(self, data):
        if self.get_iteration() == 0:
            return data[0]
        return data[0]

    def post_process(self):
        # option here to break out of the iterations
        # self.set_processing_complete()
        pass

    def setup(self):
        self.exp.log(self.name + " Start")

        # set up the output dataset that is created by the plugin
        in_dataset, out_dataset = self.get_datasets()

        in_pData, out_pData = self.get_plugin_datasets()
        in_pData[0].plugin_data_setup('SINOGRAM', 'single')

        out_dataset[0].create_dataset(in_dataset[0])
        self.clone_dataset(out_dataset[1], out_dataset[0])

        out_pData[0].plugin_data_setup('SINOGRAM', 'single')
        out_pData[1].plugin_data_setup('SINOGRAM', 'single')

        # try replacing input dataset with the output dataset
        self.set_iteration_datasets(0, [in_dataset[0]], [out_dataset[0]])
        self.set_iteration_datasets(1, [out_dataset[0]], [out_dataset[1]])

        self.exp.log(self.name + " End")

    def nOutput_datasets(self):
        return 2

    def nClone_datasets(self):
        return 1
